IMPLEMENTATION MODULE Geo;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    **************************************************************************
    Compiler:   TopSpeed / Stony Brook Modula-2
    --------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0                                           August 86
    A.Iakovlev  Ver  2.0  Procedures for rectangles                May 87
    A.Iakovlev  Ver  3.0  Trapezoids                               November 87
    A.Iakovlev  Ver  4.0  IntRectEqRect                            February 88
    A.Iakovlev  Ver  5.0  Rectangles shifting et al.               February 88
    A.Iakovlev  Ver  6.0  New hor and ver lines                    August 88
    A.Iakovlev  Ver  7.0  Hor/ver line processing                  October 88
    A.Iakovlev  Ver  7.1  Safe arithmetics                         October 88
    A.Iakovlev  Ver  8.0  Simplifications                          February 89
    A.Iakovlev  Ver  9.0  Hor/ver lines again                      March 89
    A.Iakovlev  Ver 10.0  PtEqPt, Pt2?L and RectPtHVL              July 89
    A.Iakovlev  Ver 11.0  Renaming                                 January 90
    A.Iakovlev  Ver 12.0  RectComplement and RectMinusRect         February 90
    A.Iakovlev  Ver 13.0  No hor/ver line processing               June 90
    A.Iakovlev  Ver 13.1  Options for speed                        July 90
    A.Iakovlev  Ver 14.0  Procedure ShiftXrect                     August 90
    A.Iakovlev  Ver 15.0  Procedure RectURectEnclose               September 90
    A.Iakovlev  Ver 15.1  No WITH statement                        February 91
    A.Iakovlev  Ver 16.0  Procedure TraceLine                      June 92
    A.Iakovlev  Ver 17.0  High indexes in RectComplement/MinusRect June 92
    A.Iakovlev  Ver 18.0  Renaming                                 June 93
    A.Iakovlev  Ver 18.1  Erasure of shift operator                October 93
    A.Iakovlev  Ver 19.0  Rectangle/Circle/Ellipse tracing         October 93
    A.Iakovlev  Ver 20.0  Import from SYSTEM                       January 94
    A.Iakovlev  Ver 20.1  Stony Brook support                      February 94
    **************************************************************************
    TopSpeed compiler dependencies:
    ------ options (*$...*)
    ------ structured constants
    Stony Brook compiler dependencies:
    ------ structured constants
  *)

  FROM Cfg IMPORT CMPTS, CMPSB;
  IMPORT SYSTEM;

  (*$I-,O-,S-,Z-*)

  CONST
    MinInt = MIN(INTEGER);
    MaxInt = MAX(INTEGER);

  (* ======================== Structure constants *)

  (*%T CMPTS*)
  CONST MaxRect = Rectangle(MinInt, MaxInt, MinInt, MaxInt);
  (*%E*)

  (*%T CMPSB*)
  CONST MaxRect: Rectangle = [MinInt, MaxInt, MinInt, MaxInt];
  (*%E*)

  (* ================================================== Utilities *)

  (*$N*)

  PROCEDURE min(a, b: INTEGER): INTEGER;
    BEGIN
      IF a <= b THEN
        RETURN a;
      ELSE
        RETURN b;
      END;
    END min;

  PROCEDURE max(a, b: INTEGER): INTEGER;
    BEGIN
      IF a <= b THEN
        RETURN b;
      ELSE
        RETURN a;
      END;
    END max;

  PROCEDURE Order(VAR a, b: INTEGER);
  (* -------- check and set order a<=b *)
    VAR
      c: INTEGER;
    BEGIN
      IF a > b THEN
        (* ------ swapping *)
        c := a;
        a := b;
        b := c;
      END;
    END Order;

  PROCEDURE SafeAdd(a, b: INTEGER): INTEGER;
  (* ------------------ overflow-free addition *)
    BEGIN
      IF (b > 0) AND (a > MaxInt - b) THEN
        RETURN MaxInt;
      ELSIF (b < 0) AND (a < MinInt - b) THEN
        RETURN MinInt;
      ELSE
        RETURN a + b;
      END;
    END SafeAdd;

  PROCEDURE SafeMean(a, b: INTEGER): INTEGER;
  (* ---------- overflow-free arithmetic mean *)
    VAR
      c: INTEGER;
    BEGIN
      c := a DIV 2 + b DIV 2;
      IF ODD(a) AND ODD(b) THEN
        INC(c);
      END;
      RETURN c;
    END SafeMean;

  (*$F*)

  (* ====================================== Points *)

  PROCEDURE PtEqPt(p1, p2: Point): BOOLEAN;
    BEGIN
      RETURN (p1.ix = p2.ix) AND (p1.iy = p2.iy);
    END PtEqPt;

  (* ============================================== Rectangles *)

  PROCEDURE SetRect(VAR r: Rectangle; x1, x2, y1, y2: INTEGER);
    BEGIN
      r.ixMin := x1;
      r.ixMax := x2;
      r.iyMin := y1;
      r.iyMax := y2;
      Order(r.ixMin, r.ixMax);
      Order(r.iyMin, r.iyMax);
    END SetRect;

  PROCEDURE RectNotEmpty(r: Rectangle): BOOLEAN;
    BEGIN
      RETURN (r.ixMin <= r.ixMax) AND (r.iyMin <= r.iyMax);
    END RectNotEmpty;

  PROCEDURE RectComplement(r: Rectangle;
            VAR res: ARRAY OF Rectangle; inx, high: CARDINAL): CARDINAL;
    BEGIN
      IF r.iyMin > MinInt THEN
        IF inx <= high THEN
          res[inx] := MaxRect;
          res[inx].iyMax := r.iyMin - 1;
        END;
        INC(inx);
      END;
      IF r.iyMax < MaxInt THEN
        IF inx <= high THEN
          res[inx] := MaxRect;
          res[inx].iyMin := r.iyMax + 1;
        END;
        INC(inx);
      END;
      IF r.ixMin > MinInt THEN
        IF inx <= high THEN
          res[inx] := r;
          res[inx].ixMin := MinInt;
          res[inx].ixMax := r.ixMin - 1;
        END;
        INC(inx);
      END;
      IF r.ixMax < MaxInt THEN
        IF inx <= high THEN
          res[inx] := r;
          res[inx].ixMin := r.ixMax + 1;
          res[inx].ixMax := MaxInt;
        END;
        INC(inx);
      END;
      RETURN inx;
    END RectComplement;

  PROCEDURE RectEqRect(r1, r2: Rectangle): BOOLEAN;
    BEGIN
      RETURN
        (r2.ixMin = r1.ixMin) AND (r1.ixMax = r2.ixMax) AND
        (r2.iyMin = r1.iyMin) AND (r1.iyMax = r2.iyMax);
    END RectEqRect;

  PROCEDURE RectInRect(r1, r2: Rectangle): BOOLEAN;
    BEGIN
      RETURN
        (r2.ixMin <= r1.ixMin) AND (r1.ixMax <= r2.ixMax) AND
        (r2.iyMin <= r1.iyMin) AND (r1.iyMax <= r2.iyMax);
    END RectInRect;

  PROCEDURE RectURectEnclose(r1, r2: Rectangle; VAR enclose: Rectangle);
    BEGIN
      enclose.ixMin := min(r1.ixMin, r2.ixMin);
      enclose.ixMax := max(r1.ixMax, r2.ixMax);
      enclose.iyMin := min(r1.iyMin, r2.iyMin);
      enclose.iyMax := max(r1.iyMax, r2.iyMax);
    END RectURectEnclose;

  PROCEDURE RectXRectRes(r1, r2: Rectangle; VAR inter: Rectangle): BOOLEAN;
    BEGIN
      inter.ixMin := max(r1.ixMin, r2.ixMin);
      inter.ixMax := min(r1.ixMax, r2.ixMax);
      inter.iyMin := max(r1.iyMin, r2.iyMin);
      inter.iyMax := min(r1.iyMax, r2.iyMax);
      RETURN (inter.ixMin <= inter.ixMax) AND (inter.iyMin <= inter.iyMax);
    END RectXRectRes;

  PROCEDURE RectXRect(r1, r2: Rectangle): BOOLEAN;
    BEGIN
      RETURN
        (r2.ixMax >= r1.ixMin) AND (r1.ixMax >= r2.ixMin) AND
        (r2.iyMax >= r1.iyMin) AND (r1.iyMax >= r2.iyMin);
    END RectXRect;

  PROCEDURE RectMinusRect(r1, r2: Rectangle;
            VAR res: ARRAY OF Rectangle; inx, high: CARDINAL): CARDINAL;
    CONST
      MaxNmbOfRects = 4;
    VAR
      cr: ARRAY [0 .. MaxNmbOfRects - 1] OF Rectangle;
      xr: Rectangle;
      j, n: CARDINAL;
    BEGIN
      IF RectXRect(r1, r2) THEN
        (* --------- intersection exist: compute the complement for r2 *)
        n := RectComplement(r2, cr, 0, MaxNmbOfRects - 1);
        (* -------------- compute the intersection with the complement *)
        IF n > 0 THEN
          FOR j := 0 TO n - 1 DO
            IF RectXRectRes(r1, cr[j], xr) THEN
              IF inx <= high THEN
                res[inx] := xr;
              END;
              INC(inx);
            END;
          END;
        END;
      ELSE
        (* ------------ no intersection *)
        IF inx <= high THEN
          res[inx] := r1;
        END;
        INC(inx);
      END;
      RETURN inx;
    END RectMinusRect;

  PROCEDURE ShiftRect(VAR r: Rectangle; dx, dy: INTEGER);
    BEGIN
      r.ixMin := SafeAdd(r.ixMin, dx);
      r.ixMax := SafeAdd(r.ixMax, dx);
      r.iyMin := SafeAdd(r.iyMin, dy);
      r.iyMax := SafeAdd(r.iyMax, dy);
    END ShiftRect;

  PROCEDURE ShiftXRect(VAR r: Rectangle; dx, dy: INTEGER): BOOLEAN;
    BEGIN
      r.ixMin := max(r.ixMin, SafeAdd(r.ixMin, dx));
      r.ixMax := min(r.ixMax, SafeAdd(r.ixMax, dx));
      r.iyMin := max(r.iyMin, SafeAdd(r.iyMin, dy));
      r.iyMax := min(r.iyMax, SafeAdd(r.iyMax, dy));
      RETURN (r.ixMin <= r.ixMax) AND (r.iyMin <= r.iyMax);
    END ShiftXRect;

  (* ================================== Rectangles and Points *)

  PROCEDURE SetRectPt(VAR r: Rectangle; p1, p2: Point);
    BEGIN
      r.ixMin := p1.ix;
      r.iyMin := p1.iy;
      r.ixMax := p2.ix;
      r.iyMax := p2.iy;
      Order(r.ixMin, r.ixMax);
      Order(r.iyMin, r.iyMax);
    END SetRectPt;

  PROCEDURE RectBasePt(r: Rectangle; VAR p: Point);
    BEGIN
      p.ix := r.ixMin;
      p.iy := r.iyMin;
    END RectBasePt;

  PROCEDURE RectCentrePt(r: Rectangle; VAR p: Point);
    BEGIN
      p.ix := SafeMean(r.ixMin, r.ixMax);
      p.iy := SafeMean(r.iyMin, r.iyMax);
    END RectCentrePt;

  PROCEDURE PtInRect(p: Point; r: Rectangle): BOOLEAN;
    BEGIN
      RETURN
        (r.ixMin <= p.ix) AND (p.ix <= r.ixMax) AND
        (r.iyMin <= p.iy) AND (p.iy <= r.iyMax);
    END PtInRect;

  (* ============================================= Tracing by rectangles *)

  PROCEDURE TraceLine(p1, p2: Point; proc: Action; data: SYSTEM.ADDRESS);
  (* ---------------- modified Bresenham's algorithm *)
    VAR
      dx, dy, sx, sy, dminus, dplus, d: INTEGER;
      r: Rectangle;
    BEGIN
      (* -------------------------- solid line *)
      dx := p2.ix - p1.ix;
      dy := p2.iy - p1.iy;
      (* ---------------------------------------- set rectangle *)
      r.ixMin := p1.ix;
      r.ixMax := p1.ix;
      r.iyMin := p1.iy;
      r.iyMax := p1.iy;
      (* -------------------------------------------- select quadrant *)
      IF dx - dy > 0 THEN
        IF dx + dy > 0 THEN
          (* ---------------- quadrant dx - dy > 0, dx + dy > 0 *)
          IF dy > 0 THEN sy := 1 ELSE sy := - 1 END;
          dminus := ABS(dy) * 2;
          d := dminus - dx;
          dplus := d - dx;
          (* -------------------------- main cycle *)
          WHILE r.ixMax # p2.ix DO
            IF d < 0 THEN
              INC(r.ixMax);
              INC(d, dminus);
            ELSE
              proc(r, data);
              INC(r.ixMax);
              r.ixMin := r.ixMax;
              INC(r.iyMin, sy);
              INC(r.iyMax, sy);
              INC(d, dplus);
            END;
          END;
          (* ------------------ last line *)
          proc(r, data);
        ELSE
          (* --------------- quadrant dx - dy > 0, dx + dy <= 0 *)
          IF dx > 0 THEN sx := 1 ELSE sx := - 1 END;
          dminus := ABS(dx) * 2;
          d := dminus + dy;
          dplus := d + dy;
          (* --------------------------- main cycle *)
          WHILE r.iyMin # p2.iy DO
            IF d < 0 THEN
              DEC(r.iyMin);
              INC(d, dminus);
            ELSE
              proc(r, data);
              DEC(r.iyMin);
              r.iyMax := r.iyMin;
              INC(r.ixMin, sx);
              INC(r.ixMax, sx);
              INC(d, dplus);
            END;
          END;
          (* ------------------ last line *)
          proc(r, data);
        END;
      ELSE
        IF dx + dy > 0 THEN
          (* ------------- quadrant dx - dy <= 0, dx + dy > 0 *)
          IF dx > 0 THEN sx := 1 ELSE sx := - 1 END;
          dminus := ABS(dx) * 2;
          d := dminus - dy;
          dplus := d - dy;
          (* --------------------------- main cycle *)
          WHILE r.iyMax # p2.iy DO
            IF d < 0 THEN
              INC(r.iyMax);
              INC(d, dminus);
            ELSE
              proc(r, data);
              INC(r.iyMax);
              r.iyMin := r.iyMax;
              INC(r.ixMin, sx);
              INC(r.ixMax, sx);
              INC(d, dplus);
            END;
          END;
          (* ------------------ last line *)
          proc(r, data);
        ELSE
          (* ---------------- quadrant dx - dy <= 0, dx + dy <= 0 *)
          IF dy > 0 THEN sy := 1 ELSE sy := - 1 END;
          dminus := ABS(dy) * 2;
          d := dminus + dx;
          dplus := d + dx;
          (* --------------------------- main cycle *)
          WHILE r.ixMin # p2.ix DO
            IF d < 0 THEN
              DEC(r.ixMin);
              INC(d, dminus);
            ELSE
              proc(r, data);
              DEC(r.ixMin);
              r.ixMax := r.ixMin;
              INC(r.iyMin, sy);
              INC(r.iyMax, sy);
              INC(d, dplus);
            END;
          END;
          (* ------------------ last line *)
          proc(r, data);
        END;
      END;
    END TraceLine;

  PROCEDURE TraceRect(r: Rectangle; proc: Action; data: SYSTEM.ADDRESS);
    VAR
      s, t: INTEGER;
    BEGIN
      (* ----------------------------- check rectangle internity *)
      IF (r.iyMin < r.iyMax - 1) AND (r.ixMin < r.ixMax - 1) THEN
        (* ---------------------------------------------- top side *)
        s := r.iyMin;
        t := r.iyMax;
        r.iyMax := s;
        proc(r, data);
        (* ---------------------------------------------- bottom side *)
        r.iyMin := t;
        r.iyMax := t;
        proc(r, data);
        (* -------------------------------------------- left side *)
        r.iyMin := s + 1;
        DEC(r.iyMax);
        t := r.ixMax;
        r.ixMax := r.ixMin;
        proc(r, data);
        (* -------------------------------------- right side *)
        r.ixMin := t;
        r.ixMax := t;
        proc(r, data);
      ELSE
        (* ----------------- rectangle has no internity *)
        proc(r, data);
      END;
    END TraceRect;

  PROCEDURE TraceCircle(p: Point; radius: CARDINAL; proc: Action; data: SYSTEM.ADDRESS);
  (* ------------------------ modified Gutknecht's algorithm *)
    VAR
      x, y, q, s: INTEGER;
      r: Rectangle;
    BEGIN
      (* ------------- initialize the controls *)
      x := 0;
      y := radius;
      q := 4 * radius;
      s := 0;
      (* ----------------------------------- main loop *)
      LOOP
        (* --------- increment x-coordinate while possible *)
        REPEAT
          INC(x);
          DEC(q, 8 * x - 4);
        UNTIL q < 1;
        DEC(x);
        (* ---------------------------------------- draw points *)
        IF s = 0 THEN
          (* --------------------------------------------- on axes *)
          r.ixMin := p.ix - x;
          r.ixMax := p.ix + x;
          r.iyMin := p.iy + y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.iyMin := p.iy - y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.ixMin := p.ix + y;
          r.ixMax := r.ixMin;
          r.iyMin := p.iy - x;
          r.iyMax := p.iy + x;
          proc(r, data);
          r.ixMin := p.ix - y;
          r.ixMax := r.ixMin;
          proc(r, data);
        ELSIF x < y THEN
          (* ----------------------------------------------- internally *)
          r.ixMin := p.ix + s;
          r.ixMax := p.ix + x;
          r.iyMin := p.iy + y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.ixMin := p.ix - x;
          r.ixMax := p.ix - s;
          proc(r, data);
          r.iyMin := p.iy - y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.ixMin := p.ix + s;
          r.ixMax := p.ix + x;
          proc(r, data);
          r.ixMin := p.ix + y;
          r.ixMax := r.ixMin;
          r.iyMin := p.iy - x;
          r.iyMax := p.iy - s;
          proc(r, data);
          r.iyMin := p.iy + s;
          r.iyMax := p.iy + x;
          proc(r, data);
          r.ixMin := p.ix - y;
          r.ixMax := r.ixMin;
          proc(r, data);
          r.iyMin := p.iy - x;
          r.iyMax := p.iy - s;
          proc(r, data);
        ELSIF x = y THEN
          (* --------------------- on bissectrice, and exit *)
          r.ixMin := p.ix + x;
          r.ixMax := r.ixMin;
          r.iyMin := p.iy + y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.iyMin := p.iy - y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.ixMin := p.ix - x;
          r.ixMax := r.ixMin;
          proc(r, data);
          r.iyMin := p.iy + y;
          r.iyMax := r.iyMin;
          proc(r, data);
          EXIT;
        ELSE
          (* ----------------------------- only exit *)
          EXIT;
        END;
        (* ------------- decrement y-coordinate *)
        INC(x);
        DEC(y);
        INC(q, 8 * y);
        s := x;
      END;
    END TraceCircle;

  PROCEDURE TraceFillCircle(p: Point; radius: CARDINAL; proc: Action; data: SYSTEM.ADDRESS);
  (* ----------------------- modified Gutknecht's algorithm *)
    VAR
      x, y, q, s: INTEGER;
      r: Rectangle;
    BEGIN
      (* --------------- initialize the controls *)
      x := 0;
      y := radius;
      q := 4 * radius;
      s := 0;
      (* --------------------------------- main loop *)
      LOOP
        (* ------ increment x-coordinate while possible *)
        REPEAT
          INC(x);
          DEC(q, 8 * x - 4);
        UNTIL q < 1;
        DEC(x);
        (* -------------------------------------- draw points *)
        IF s = 0 THEN
          (* ------------------------------------------ on axes *)
          r.ixMin := p.ix - x;
          r.ixMax := p.ix + x;
          r.iyMin := p.iy + y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.iyMin := p.iy - y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.ixMin := p.ix - y;
          r.ixMax := p.ix + y;
          r.iyMin := p.iy - x;
          r.iyMax := p.iy + x;
          proc(r, data);
        ELSIF x < y THEN
          (* --------------------------------------------- internally *)
          r.ixMin := p.ix - x;
          r.ixMax := p.ix + x;
          r.iyMin := p.iy + y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.iyMin := p.iy - y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.ixMin := p.ix - y;
          r.ixMax := p.ix + y;
          r.iyMin := p.iy + s;
          r.iyMax := p.iy + x;
          proc(r, data);
          r.iyMin := p.iy - x;
          r.iyMax := p.iy - s;
          proc(r, data);
        ELSIF x = y THEN
          (* ------------------------ on bissectrice, and exit *)
          r.ixMin := p.ix - x;
          r.ixMax := p.ix + x;
          r.iyMin := p.iy + s;
          r.iyMax := p.iy + y;
          proc(r, data);
          r.iyMin := p.iy - y;
          r.iyMax := p.iy - s;
          proc(r, data);
          EXIT;
        ELSE
          (* ------------------------------ only exit *)
          EXIT;
        END;
        (* --------------- decrement y-coordinate *)
        INC(x);
        DEC(y);
        INC(q, 8 * y);
        s := x;
      END;
    END TraceFillCircle;

  PROCEDURE TraceEllipse(p: Point; a0, b0: CARDINAL; proc: Action; data: SYSTEM.ADDRESS);
  (* ----------------- modified algorithm from the article of Van Aken & Novak,
                       but without division by 2,
                       and drawing is done from y-axis to x-axis *)
    VAR
      x, y, s: INTEGER;
      a1, b1, asq, asq2, bsq, bsq2, d1, d2, dx, dy: LONGINT;
      r: Rectangle;
    BEGIN
      (* -------------- initialize the controls *)
      x := 0;
      y := b0;
      s := 0;
      a1 := VAL(LONGINT, a0);
      b1 := VAL(LONGINT, b0);
      asq := a1 * a1;
      asq2 := asq * 2;
      bsq := b1 * b1;
      bsq2 := bsq * 2;
      d1 := (asq DIV 4) - (asq * b1) + bsq;
      d2 := asq - (2 * asq * b1) + (bsq DIV 4);
      dx := 0;
      dy := asq2 * b1;
      (* ------------------------ arc near the y-axis *)
      WHILE d2 < 0 DO
        (* -------- increment x-coordinate while possible *)
        LOOP
          INC(x);
          INC(dx, bsq2);
          IF (d1 >= 0) OR (d2 >= 0) THEN EXIT END;
          INC(d1, bsq + dx);
          INC(d2, dx);
        END;
        DEC(x);
        (* --------------------------------------- draw points *)
        IF s = 0 THEN
          (* --------------------------- with maximum y-coordinate *)
          r.ixMin := p.ix - x;
          r.ixMax := p.ix + x;
          r.iyMin := p.iy + y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.iyMin := p.iy - y;
          r.iyMax := r.iyMin;
          proc(r, data);
        ELSE
          (* --------------------------------------------- internally *)
          r.ixMin := p.ix + s;
          r.ixMax := p.ix + x;
          r.iyMin := p.iy + y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.ixMin := p.ix - x;
          r.ixMax := p.ix - s;
          proc(r, data);
          r.iyMin := p.iy - y;
          r.iyMax := r.iyMin;
          proc(r, data);
          r.ixMin := p.ix + s;
          r.ixMax := p.ix + x;
          proc(r, data);
        END;
        (* ---------------------------------- decrement y-coordinate *)
        INC(x);
        DEC(y);
        DEC(dy, asq2);
        INC(d1, bsq + dx - dy);
        INC(d2, asq + dx - dy);
        s := x;
      END;
      (* ----------------------------------- arc near the x-axis *)
      s := y;
      LOOP
        (* --------------- decrement y-coordinate while possible *)
        LOOP
          DEC(y);
          DEC(dy, asq2);
          IF (d2 < 0) OR (y < 0) THEN EXIT END;
          INC(d2, asq - dy);
        END;
        INC(y);
        (* -------------------------------------- draw points *)
        IF y > 0 THEN
          (* ----------------------------------- internally *)
          r.ixMin := p.ix + x;
          r.ixMax := r.ixMin;
          r.iyMin := p.iy + y;
          r.iyMax := p.iy + s;
          proc(r, data);
          r.ixMin := p.ix - x;
          r.ixMax := r.ixMin;
          proc(r, data);
          r.iyMin := p.iy - s;
          r.iyMax := p.iy - y;
          proc(r, data);
          r.ixMin := p.ix + x;
          r.ixMax := r.ixMin;
          proc(r, data);
        ELSE
          (* ------------------------------ on x-axis *)
          r.ixMin := p.ix + x;
          r.ixMax := r.ixMin;
          r.iyMin := p.iy - s;
          r.iyMax := p.iy + s;
          proc(r, data);
          r.ixMin := p.ix - x;
          r.ixMax := r.ixMin;
          proc(r, data);
          EXIT;
        END;
        (* ------------- increment x-coordinate *)
        DEC(y);
        INC(x);
        INC(dx, bsq2);
        INC(d2, asq + dx - dy);
        s := y;
      END;
    END TraceEllipse;

  PROCEDURE TraceFillEllipse(p: Point; a0, b0: CARDINAL; proc: Action; data: SYSTEM.ADDRESS);
  (* ----------------- modified algorithm from the article of Van Aken & Novak,
                       but without division by 2,
                       and drawing is done from y-axis to x-axis *)
    VAR
      x, y, s: INTEGER;
      a1, b1, asq, asq2, bsq, bsq2, d1, d2, dx, dy: LONGINT;
      r: Rectangle;
    BEGIN
      (* --------- initialize the controls *)
      x := 0;
      y := b0;
      a1 := VAL(LONGINT, a0);
      b1 := VAL(LONGINT, b0);
      asq := a1 * a1;
      asq2 := asq * 2;
      bsq := b1 * b1;
      bsq2 := bsq * 2;
      d1 := (asq DIV 4) - (asq * b1) + bsq;
      d2 := asq - (2 * asq * b1) + (bsq DIV 4);
      dx := 0;
      dy := asq2 * b1;
      (* --------------------- arc near the y-axis *)
      WHILE d2 < 0 DO
        (* -------- increment x-coordinate while possible *)
        LOOP
          INC(x);
          INC(dx, bsq2);
          IF (d1 >= 0) OR (d2 >= 0) THEN EXIT END;
          INC(d1, bsq + dx);
          INC(d2, dx);
        END;
        DEC(x);
        (* ------------------------------------------- draw points *)
        r.ixMin := p.ix - x;
        r.ixMax := p.ix + x;
        r.iyMin := p.iy + y;
        r.iyMax := r.iyMin;
        proc(r, data);
        r.iyMin := p.iy - y;
        r.iyMax := r.iyMin;
        proc(r, data);
        (* ---------------------------------------- decrement y-coordinate *)
        INC(x);
        DEC(y);
        DEC(dy, asq2);
        INC(d1, bsq + dx - dy);
        INC(d2, asq + dx - dy);
      END;
      (* --------------------------------------------- arc near the x-axis *)
      s := y;
      LOOP
        (* ------------------------ decrement y-coordinate while possible *)
        LOOP
          DEC(y);
          DEC(dy, asq2);
          IF (y < 0) OR (d2 < 0) THEN EXIT END;
          INC(d2, asq - dy);
        END;
        INC(y);
        (* ------------------------------------------- draw points *)
        IF y > 0 THEN
          (* ---------------------------------------- internally *)
          r.ixMin := p.ix - x;
          r.ixMax := p.ix + x;
          r.iyMin := p.iy + y;
          r.iyMax := p.iy + s;
          proc(r, data);
          r.iyMin := p.iy - s;
          r.iyMax := p.iy - y;
          proc(r, data);
        ELSE
          (* ------------------------------------- on x-axis *)
          r.ixMin := p.ix - x;
          r.ixMax := p.ix + x;
          r.iyMin := p.iy - s;
          r.iyMax := p.iy + s;
          proc(r, data);
          EXIT;
        END;
        (* ----------------- increment x-coordinate *)
        DEC(y);
        INC(x);
        INC(dx, bsq2);
        INC(d2, asq + dx - dy);
        s := y;
      END;
    END TraceFillEllipse;

  END Geo.
