IMPLEMENTATION MODULE ScrBar;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    ************************************************************************
    Compilers:  TopSpeed / Stony Brook Modula-2
    ------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Initial version                        October 94
    A.Iakovlev  Ver  1.1  Correction of error in Init            March 95
    A.Iakovlev  Ver  2.0  Simplifications in Create/Destroy      March 95
    ************************************************************************
    TopSpeed compiler dependencies:
    ------ predefined type LONGCARD
    Stony Brook compiler dependencies:
    ------ predefined type LONGCARD
  *)

  IMPORT SYSTEM, DrvMem, ScrTyp, ScrExt;

  PROCEDURE Create(pm: ScrTyp.PIXMAP): BOOLEAN;
    VAR
      bar, size: CARDINAL;
      alloc: BOOLEAN;
    BEGIN
      WITH pm^ DO
        IF DrvMem.Alloc(cont.addr, (number + 1) * SIZE(SYSTEM.ADDRESS)) THEN
          size := BytesPerBar;
          bar := 0;
          (* --------------------- attempt to allocate all the bars *)
          REPEAT
            IF bar = number THEN size := length END;
            alloc := DrvMem.Alloc(cont.dir^[bar], size);
            INC(bar);
          UNTIL (NOT alloc) OR (bar = number + 1);
          (* ---------------------------- test and possible memory clearing *)
          IF NOT alloc THEN
            WHILE bar > 0 DO
              DEC(bar);
              IF bar # number THEN size := BytesPerBar END;
              DrvMem.Free(cont.dir^[bar], size);
            END;
            DrvMem.Free(cont.addr, (number + 1) * SIZE(SYSTEM.ADDRESS));
          END;
          (* ---------------------- normal return *)
          RETURN alloc;
        ELSE
          RETURN FALSE;
        END;
      END;
    END Create;

  PROCEDURE Init(pm: ScrTyp.PIXMAP; ix, iy: INTEGER; VAR bp: BarPtr);
    VAR
      ds: ScrTyp.BP;
    BEGIN
      bp.pm := pm;
      WITH bp.pm^ DO
        IF screen THEN
          (* ---------------------------- screen pixmap *)
          bp.ofs := 0;
          bp.ptr.long :=
            VAL(LONGCARD, iy) * VAL(LONGCARD, width) +
            VAL(LONGCARD, ix) * VAL(LONGCARD, bytesPerPixel);
          bp.bar := bp.ptr.seg;
          bp.ptr.seg := cont.seg;
          ScrExt.SetBank(bp.bar);
        ELSIF number > 0 THEN
          (* ---------------------------------- multi-bar memory pixmap *)
          ds.long :=
            VAL(LONGCARD, iy - iy0) * VAL(LONGCARD, width) +
            VAL(LONGCARD, ix - ix0) * VAL(LONGCARD, bytesPerPixel);
          bp.bar := VAL(CARDINAL, ds.long DIV BytesPerBar);
          bp.ptr := ScrTyp.BP(cont.dir^[bp.bar]);
          bp.ofs := bp.ptr.ofs;
          INC(bp.ptr.ofs, VAL(CARDINAL, ds.long MOD BytesPerBar));
        ELSE
          (* -------------------------- single-bar memory pixmap *)
          bp.bar := 0;
          bp.ptr := cont;
          bp.ofs := bp.ptr.ofs;
          INC(bp.ptr.ofs,
            VAL(CARDINAL, iy - iy0) * width +
            VAL(CARDINAL, ix - ix0) * bytesPerPixel);
        END;
      END;
    END Init;

  PROCEDURE Ofs(bp: BarPtr): CARDINAL;
    BEGIN
      RETURN bp.ptr.ofs - bp.ofs;
    END Ofs;

  PROCEDURE Inc(VAR bp: BarPtr; delta: CARDINAL);
    VAR
      oldOfs: CARDINAL;
    BEGIN
      WITH bp.pm^ DO
        IF screen THEN
          (* ---------------------------------- screen pixmap *)
          IF delta <= barLength - bp.ptr.ofs THEN
            INC(bp.ptr.ofs, delta);
          ELSE
            INC(bp.bar);
            bp.ptr.ofs := delta - (barLength - bp.ptr.ofs) - 1;
            ScrExt.SetBank(bp.bar);
          END;
        ELSIF number > 0 THEN
          (* ----------------------------------- multi-bar memory pixmap *)
          IF delta <= barLength - (bp.ptr.ofs - bp.ofs) THEN
            INC(bp.ptr.ofs, delta);
          ELSE
            INC(bp.bar);
            oldOfs := bp.ptr.ofs - bp.ofs;
            bp.ptr := ScrTyp.BP(cont.dir^[bp.bar]);
            bp.ofs := bp.ptr.ofs;
            INC(bp.ptr.ofs, delta - (barLength - oldOfs) - 1);
          END;
        ELSE
          (* -------------------------- single-bar memory pixmap *)
          INC(bp.ptr.ofs, delta);
        END;
      END;
    END Inc;

  PROCEDURE Dec(VAR bp: BarPtr; delta: CARDINAL);
    VAR
      oldOfs: CARDINAL;
    BEGIN
      WITH bp.pm^ DO
        IF screen THEN
          (* ---------------------------------- screen pixmap *)
          IF delta <= bp.ptr.ofs THEN
            DEC(bp.ptr.ofs, delta);
          ELSE
            DEC(bp.bar);
            bp.ptr.ofs := barLength - (delta - bp.ptr.ofs - 1);
            ScrExt.SetBank(bp.bar);
          END;
        ELSIF number > 0 THEN
          (* ----------------------------------- multi-bar memory pixmap *)
          IF delta <= bp.ptr.ofs - bp.ofs THEN
            DEC(bp.ptr.ofs, delta);
          ELSE
            DEC(bp.bar);
            oldOfs := bp.ptr.ofs - bp.ofs;
            bp.ptr := ScrTyp.BP(cont.dir^[bp.bar]);
            bp.ofs := bp.ptr.ofs;
            INC(bp.ptr.ofs, barLength - (delta - oldOfs - 1));
          END;
        ELSE
          (* -------------------------- single-bar memory pixmap *)
          DEC(bp.ptr.ofs, delta);
        END;
      END;
    END Dec;

  PROCEDURE Destroy(pm: ScrTyp.PIXMAP);
    VAR
      bar, size: CARDINAL;
    BEGIN
      WITH pm^ DO
        size := length;
        FOR bar := number TO 0 BY - 1 DO
          IF bar # number THEN size := BytesPerBar END;
          DrvMem.Free(cont.dir^[bar], size);
        END;
        DrvMem.Free(cont.addr, (number + 1) * SIZE(SYSTEM.ADDRESS));
      END;
    END Destroy;

  END ScrBar.
