IMPLEMENTATION MODULE ScrExt;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    ************************************************************************
    Hardware :  IBM PC/XT/AT ROM BIOS, VESA Video BIOS Extension
    OpSystem :  MS-DOS & TopSpeed DOS Extender
    Compilers:  TopSpeed / Stony Brook Modula-2
    ------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Derived from module Scr               August 92
    A.Iakovlev  Ver  2.0  Renaming                              June 93
    A.Iakovlev  Ver  3.0  Autodetection of Super-VGA hardware   June 93
    A.Iakovlev  Ver  4.0  Memory determination                  June 93
    A.Iakovlev  Ver  4.1  Stony Brook support                   February 94
    A.Iakovlev  Ver  5.0  Full redesign, VESA support           March 94
    A.Iakovlev  Ver  6.0  Video selector, font info procedure   March 94
    A.Iakovlev  Ver  6.1  Correction in PrepareSuper            April 94
    A.Iakovlev  Ver  7.0  Use Scr.Pixel                         September 94
    A.Iakovlev  Ver  8.0  High and true-color modes             September 94
    A.Iakovlev  Ver  8.1  Use module Mem for heap control       November 94
    A.Iakovlev  Ver  9.0  Variable bgr                          February 95
    A.Iakovlev  Ver  9.1  Checking of bytesPerShift value       April 95
    ************************************************************************
    MS-DOS dependencies:
    ------ video driver at INT 10H (functions 0, 5, 0FH, 11H, 12H, 4FH)
    TopSpeed DOS Extender dependencies:
    ------ necessity to initialize the interceptor for the video interrupt
    ------ import from TSXLIB: SEL_A000H, InitInt10, ALLOCLOWSEG, FREESEG,
           MAKEREALSEG, MAKESELECTOR, REALINTR, REALCALL
    ------ necessity to make the selector for absolute address
    TopSpeed compiler dependencies:
    ------ pragmas (*#...*)
    ------ predefined type LONGCARD
    ------ single-byte sets
    TopSpeed library dependencies:
    ------ import from SYSTEM: BYTE
    ------ import from Strings: CompareStr
    Stony Brook compiler dependencies:
    ------ predefined type LONGCARD
    ------ single-byte sets
    Stony Brook library dependencies:
    ------ import from SYSTEM: BYTE
    ------ import from Strings: CompareStr
  *)

  FROM Cfg IMPORT CMPTS, CMPSB, REGNO, REGTS, EXTNO, EXTTS;
  IMPORT
    SYSTEM, Strings,
    DrvMem, SrvSys, SrvLow, ScrTyp, ScrLow, ScrBas, Scr;

  (* =============================== Conditional import *)

  (*%T EXTTS*)
  IMPORT TSXLIB;
  (*%E*)

  (* ========================================= Global constants *)

  CONST
    ScreenSeg = 0A000H;
    DefaultMemory = 16; (* --- in 64 Kbytes *)
    BytesPerBank = 10000H;
    BitsPerByte = 8;

  (* ============================================ Global variables *)

  VAR
    oldMode, oldPage, biosMode, subMode: CARDINAL;
    bytesPerShift: CARDINAL;

  (* ====================================== Hardware/OpSystem interface *)

  CONST
    VideoInterrupt = 10H;
    SetModeFun = 0;
    SetPageFun = 500H;
    GetModeFun = 0F00H;
    GetFontInfoFun = 1130H;
    SpecialFun = 1200H;
    GetConfigSubFun = 10H;

  PROCEDURE GetMemory(): CARDINAL;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AX := SpecialFun;
      reg.BL := GetConfigSubFun;
      SrvSys.Int(reg, VideoInterrupt);
      RETURN ORD(reg.BL) + 1;
    END GetMemory;

  PROCEDURE StoreModeAndPage();
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AX := GetModeFun;
      SrvSys.Int(reg, VideoInterrupt);
      oldMode := ORD(reg.AL);
      oldPage := ORD(reg.BH);
    END StoreModeAndPage;

  PROCEDURE RestoreModeAndPage();
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AX := SetModeFun + oldMode;
      SrvSys.Int(reg, VideoInterrupt);
      reg.AX := SetPageFun + oldPage;
      SrvSys.Int(reg, VideoInterrupt);
    END RestoreModeAndPage;

  (* ================================================= Utilities *)

  PROCEDURE EmptySetBank(n: CARDINAL);
    BEGIN
    END EmptySetBank;

  PROCEDURE SetLinesPerPage(m: Scr.Type; memory: CARDINAL);
  (* --- set linesPerPage parameter, using bytesPerLine parameter *)
    VAR
      bytesPerMemo, bytesPerLineForAllPlanes: LONGCARD;
    BEGIN
      bytesPerMemo := VAL(LONGCARD, memory) * BytesPerBank;
      bytesPerLineForAllPlanes := VAL(LONGCARD, bytesPerLine * Profiles[m].planesPerPixel);
      linesPerPage := VAL(CARDINAL, bytesPerMemo DIV bytesPerLineForAllPlanes);
    END SetLinesPerPage;

  (* ===================================== Standard EGA/VGA *)

  PROCEDURE StandardInit();
    VAR
      reg: SrvSys.Registers;
    BEGIN
      StoreModeAndPage();
      reg.AX := biosMode;
      SrvSys.Int(reg, VideoInterrupt);
      bytesPerShift := ScrLow.BytesPerShift();
      IF bytesPerShift = 0 THEN bytesPerShift := bytesPerLine END;
    END StandardInit;

  PROCEDURE StandardSetDisplayStart(x, y: CARDINAL);
    BEGIN
      ScrLow.SetStartAddr(y * bytesPerShift);
    END StandardSetDisplayStart;

  (* ================================================ Modified VGA *)

  PROCEDURE ModifiedInit();
    VAR
      reg: SrvSys.Registers;
    BEGIN
      StoreModeAndPage();
      reg.AX := biosMode;
      SrvSys.Int(reg, VideoInterrupt);
      Scr.DisableVideo();
      ScrLow.ModifiedInit(subMode);
      bytesPerShift := ScrLow.BytesPerShift();
      IF bytesPerShift = 0 THEN bytesPerShift := bytesPerLine END;
    END ModifiedInit;

  (* ==================================== VESA Video BIOS Extension *)

  CONST
    SignLen = 4;
    OEMStrLen = 1024;
    ModesNmb = 256;
    (* ---------------------- *)
    VESACallFun = 4F00H;
    ReturnInfo = 0;
    ReturnModeInfo = 1;
    SetMode = 2;
    ReturnMode = 3;
    WindowControl = 5;
    SetGetDisplayStart = 7;
    (* ------------------- *)
    VESAOK = 4FH;

  (*%T REGNO*)
  TYPE WinFunction = LONGCARD;
  (*%E*)

  (*%T REGTS*)
  (*%T EXTNO*)
  (*# save, call(reg_param => (bx, dx)) *)
  TYPE WinFunction = PROCEDURE(CARDINAL, CARDINAL);
  (*# restore *)
  (*%E*)
  (*%T EXTTS*)
  TYPE WinFunction = ScrTyp.BP;
  (*%E*)
  (*%E*)

  TYPE
    VgaInfo =
      RECORD
        Signature    : ARRAY [0 .. SignLen - 1] OF CHAR;
        Version      : CARDINAL;
        OEMStringPtr : POINTER TO ARRAY [0 .. OEMStrLen - 1] OF CHAR;
        Capabilities : LONGCARD;
        VideoModePtr : POINTER TO ARRAY [0 .. ModesNmb - 1] OF CARDINAL;
        TotalMemory  : CARDINAL;
        Reserved     : ARRAY [1 .. 236] OF SYSTEM.BYTE;
      END;
    ModAttrs = (hardOk, extInfo, outBios, color, graphics, b5, b6, b7, b8);
    WinAttrs = (supported, readable, writeable, d3, d4, d5, d6, d7);
    WinAttrSet = SET OF WinAttrs;
    ModInfo =
      RECORD
        (* ---------------------- Mandatory information *)
        ModeAttributes     : SET OF ModAttrs;
        WinAAttributes     : WinAttrSet;
        WinBAttributes     : WinAttrSet;
        WinGranularity     : CARDINAL;
        WinSize            : CARDINAL;
        WinASegment        : CARDINAL;
        WinBSegment        : CARDINAL;
        WinFuncPtr         : WinFunction;
        BytesPerScanLine   : CARDINAL;
        (* ------------------------------ Optional information *)
        XResolution        : CARDINAL;
        YResolution        : CARDINAL;
        XCharSize          : SYSTEM.BYTE;
        YCharSize          : SYSTEM.BYTE;
        NumberOfPlanes     : SYSTEM.BYTE;
        BitsPerPixel       : SYSTEM.BYTE;
        NumberOfBanks      : SYSTEM.BYTE;
        MemoryModel        : SYSTEM.BYTE;
        BankSize           : SYSTEM.BYTE;
        NumberOfImagePages : SYSTEM.BYTE;
        Reserved1          : SYSTEM.BYTE;
        RedMaskSize        : SYSTEM.BYTE;
        RedFieldPosition   : SYSTEM.BYTE;
        GreenMaskSize      : SYSTEM.BYTE;
        GreenFieldPosition : SYSTEM.BYTE;
        BlueMaskSize       : SYSTEM.BYTE;
        BlueFieldPosition  : SYSTEM.BYTE;
        RsvdMaskSize       : SYSTEM.BYTE;
        RsvdFieldPosition  : SYSTEM.BYTE;
        DirectColorModeInfo: SYSTEM.BYTE;
        Reserved2          : ARRAY [1 .. 216] OF SYSTEM.BYTE;
      END;

  VAR
    win, gransPerWin: CARDINAL;
    wf: WinFunction;

  PROCEDURE SetWindowParameters(mi: ModInfo): BOOLEAN;
    CONST
      Full = WinAttrSet{supported, readable, writeable};
      WinSizeOK = 40H;
    VAR
      ok: BOOLEAN;
    BEGIN
      (* -------------------- choose the correct window *)
      IF mi.WinAAttributes >= Full THEN
        win := 0;
        ok := (mi.WinASegment = ScreenSeg);
      ELSIF mi.WinBAttributes >= Full THEN
        win := 1;
        ok := (mi.WinBSegment = ScreenSeg);
      ELSE
        ok := FALSE;
      END;
      (* ----------------------------- set other window parameters *)
      gransPerWin := WinSizeOK DIV mi.WinGranularity;
      wf := mi.WinFuncPtr;
      (* ------------------ check other parameters and return *)
      ok := ok AND (mi.WinSize = WinSizeOK);
      RETURN ok;
    END SetWindowParameters;

  (*%T EXTNO*)

  PROCEDURE Allocate(VAR addr: SYSTEM.ADDRESS; size: CARDINAL): BOOLEAN;
    BEGIN
      RETURN DrvMem.Alloc(addr, size);
    END Allocate;

  PROCEDURE Deallocate(addr: SYSTEM.ADDRESS; size: CARDINAL);
    BEGIN
      DrvMem.Free(addr, size);
    END Deallocate;

  PROCEDURE VideoInt(ax, cx: CARDINAL; addr: SYSTEM.ADDRESS): BOOLEAN;
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AX := ax;
      reg.CX := cx;
      SrvSys.Split(addr, reg.ES, reg.DI);
      SrvSys.Int(reg, VideoInterrupt);
      RETURN reg.AX = VESAOK;
    END VideoInt;

  (*%E*)

  (*%T EXTTS*)

  PROCEDURE Allocate(VAR addr: SYSTEM.ADDRESS; size: CARDINAL): BOOLEAN;
    VAR
      sel: CARDINAL;
    BEGIN
      sel := TSXLIB.ALLOCLOWSEG(size);
      addr := SrvSys.Addr(sel, 0);
      IF addr # NIL THEN
        SrvLow.Fill(addr, size, 0);
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END Allocate;

  PROCEDURE Deallocate(addr: SYSTEM.ADDRESS; size: CARDINAL);
    VAR
      sel, ofs: CARDINAL;
    BEGIN
      SrvSys.Split(addr, sel, ofs);
      TSXLIB.FREESEG(sel);
    END Deallocate;

  PROCEDURE VideoInt(ax, cx: CARDINAL; addr: SYSTEM.ADDRESS): BOOLEAN;
    VAR
      reg: SrvSys.Registers;
      sel, ofs: CARDINAL;
    BEGIN
      reg.AX := ax;
      reg.CX := cx;
      SrvSys.Split(addr, sel, ofs);
      reg.ES := TSXLIB.MAKEREALSEG(sel);
      reg.DI := 0;
      TSXLIB.REALINTR(SYSTEM.ADR(reg), VideoInterrupt);
      RETURN reg.AX = VESAOK;
    END VideoInt;

  (*%E*)

  PROCEDURE VESAInit();
    VAR
      reg: SrvSys.Registers;
    BEGIN
      (* -------------------------- get old mode *)
      reg.AX := VESACallFun + ReturnMode;
      SrvSys.Int(reg, VideoInterrupt);
      oldMode := reg.BX;
      (* ------------------------------------ set new mode *)
      reg.AX := VESACallFun + SetMode;
      reg.BX := biosMode;
      SrvSys.Int(reg, VideoInterrupt);
    END VESAInit;

  PROCEDURE VESATerm();
    VAR
      reg: SrvSys.Registers;
    BEGIN
      (* ------------------------- set old mode *)
      reg.AX := VESACallFun + SetMode;
      reg.BX := oldMode;
      SrvSys.Int(reg, VideoInterrupt);
      IF reg.AX = 0 THEN RETURN END;
      (* --------------- some BIOS'es can't set "foreign" mode numbers *)
      reg.AX := SetModeFun + oldMode;
      SrvSys.Int(reg, VideoInterrupt);
    END VESATerm;

  PROCEDURE VESASetBank(n: CARDINAL);
    VAR
      reg: SrvSys.Registers;
    BEGIN
  (*%T REGNO*)
      reg.AX := VESACallFun + WindowControl;
      reg.BX := win;
      reg.DX := n * gransPerWin;
      SrvSys.Int(reg, VideoInterrupt);
  (*%E*)
  (*%T REGTS*)
  (*%T EXTNO*)
      wf(win, n * gransPerWin);
  (*%E*)
  (*%T EXTTS*)
      reg.BX := win;
      reg.DX := n * gransPerWin;
      TSXLIB.REALCALL(SYSTEM.ADR(reg), wf.seg, wf.ofs);
  (*%E*)
  (*%E*)
    END VESASetBank;

  PROCEDURE VESASetDisplayStart(x, y: CARDINAL);
    VAR
      reg: SrvSys.Registers;
    BEGIN
      reg.AX := VESACallFun + SetGetDisplayStart;
      reg.BX := 0;
      reg.CX := x;
      reg.DX := y;
      SrvSys.Int(reg, VideoInterrupt);
    END VESASetDisplayStart;

  (* =============================================== Font procedures *)

  TYPE Parameters = ARRAY [0 .. MaxNmbOfStdFonts - 1] OF CARDINAL;

  (*%T CMPTS*)
  CONST FontCode = Parameters( 2,  3,  6);
  CONST CharHgt  = Parameters(14,  8, 16);
  (*%E*)

  (*%T CMPSB*)
  CONST FontCode:Parameters = [ 2,  3,  6];
  CONST CharHgt :Parameters = [14,  8, 16];
  (*%E*)

  PROCEDURE FontInfo(n: CARDINAL; VAR w, h: CARDINAL; VAR adr: SYSTEM.ADDRESS);
    CONST
      StdCharWdt = 8;
      StdCharHgt = 8;
      NmbOfChars = 256;
      StdByteCharWdt = (StdCharWdt - 1) DIV BitsPerByte + 1;
      StdFontLen = StdByteCharWdt * StdCharHgt * NmbOfChars;
    VAR
      reg: SrvSys.Registers;
      sel, lim: CARDINAL;
      pos: LONGCARD;
    BEGIN
      (* --------------------- requesting EGA/VGA fonts *)
      w := StdCharWdt;
      h := CharHgt[n];
  (*%T EXTNO*)
      adr := SrvLow.FontAddr(FontCode[n]);
  (*%E*)
  (*%T EXTTS*)
      reg.AX := GetFontInfoFun;
      reg.BH := SYSTEM.BYTE(FontCode[n]);
      TSXLIB.REALINTR(SYSTEM.ADR(reg), VideoInterrupt);
      pos := VAL(LONGCARD, reg.ES) * 10H + VAL(LONGCARD, reg.BP);
      lim := StdByteCharWdt * NmbOfChars * h;
      sel := TSXLIB.MAKESELECTOR(pos, lim);
      adr := SrvSys.Addr(sel, 0);
  (*%E*)
    END FontInfo;

  (* ========================================= Module preparing *)

  PROCEDURE PrepareStandard(m: Scr.Type);
    VAR
      pixelsPerByte, memSize: CARDINAL;
    BEGIN
      (* -------------- set procedure interface *)
      Init := StandardInit;
      Term := RestoreModeAndPage;
      SetBank := EmptySetBank;
      SetDisplayStart := StandardSetDisplayStart;
      (* ------------------------------------ set mode parameter *)
      biosMode := Profiles[m].biosMode;
      (* ---------------------------------- set bytesPerLine parameter *)
      pixelsPerByte := BitsPerByte DIV Profiles[m].bitsPerPixel;
      bytesPerLine := Profiles[m].pixelsPerLine DIV pixelsPerByte;
      (* ---------------------------------- set linesPerPage parameter *)
      memSize := GetMemory();
      IF biosMode = 13H THEN memSize := 1 END;
      SetLinesPerPage(m, memSize);
      (* --------------------------------- set testPassed parameter *)
      testPassed := (linesPerPage >= Profiles[m].linesPerScreen);
    END PrepareStandard;

  PROCEDURE PrepareModified(m: Scr.Type);
    VAR
      pixelsPerByte: CARDINAL;
    BEGIN
      (* -------------- set procedure interface *)
      Init := ModifiedInit;
      Term := RestoreModeAndPage;
      SetBank := EmptySetBank;
      SetDisplayStart := StandardSetDisplayStart;
      (* ------------------------------------ set mode parameters *)
      biosMode := Profiles[m].biosMode;
      subMode := Profiles[m].subMode;
      (* ---------------------------------- set bytesPerLine parameter *)
      pixelsPerByte := BitsPerByte DIV Profiles[m].bitsPerPixel;
      bytesPerLine := Profiles[m].pixelsPerLine DIV pixelsPerByte;
      (* ---------------------------------- set linesPerPage parameter *)
      SetLinesPerPage(m, GetMemory());
      (* ---------------------------------- set testPassed parameter *)
      testPassed :=
        (linesPerPage >= Profiles[m].linesPerScreen) AND
        (ScrBas.adapter = ScrBas.vga);
    END PrepareModified;

  PROCEDURE PrepareSuper(m: Scr.Type);
    CONST
      Id = "VESA";
    VAR
      vgaPtr: POINTER TO VgaInfo;
      modPtr: POINTER TO ModInfo;
      memSize: CARDINAL;
      ok: BOOLEAN;
    BEGIN
      (* ------------------------ is VESA supported? *)
      IF
        Allocate(vgaPtr, SIZE(VgaInfo)) AND
        VideoInt(VESACallFun + ReturnInfo, 0, vgaPtr) AND
        (Strings.CompareStr(vgaPtr^.Signature, Id) = 0)
      THEN
        (* ---------------------------- yes: set mode parameter *)
        biosMode := Profiles[m].biosMode;
        (* -------------------------------- is VESA mode supported *)
        IF
          Allocate(modPtr, SIZE(ModInfo)) AND
          VideoInt(VESACallFun + ReturnModeInfo, biosMode, modPtr) AND
          (hardOk IN modPtr^.ModeAttributes) AND
          SetWindowParameters(modPtr^)
        THEN
          (* ----------------------------- yes: set procedure interface *)
          Init := VESAInit;
          Term := VESATerm;
          SetBank := VESASetBank;
          SetDisplayStart := VESASetDisplayStart;
          (* ---------------------------------- set number of bytes per line *)
          bytesPerLine := modPtr^.BytesPerScanLine;
          (* --------------------------------- set number of lines per page *)
          IF m = Scr.mode4x800x600 THEN
            memSize := GetMemory();
          ELSE
            memSize := vgaPtr^.TotalMemory;
            IF memSize > 64 THEN memSize := DefaultMemory END;
          END;
          SetLinesPerPage(m, memSize);
          (* --------------------------------------------- set rgb order *)
          IF ORD(modPtr^.BitsPerPixel) = 24 THEN
            bgr := ORD(modPtr^.BlueFieldPosition) < ORD(modPtr^.RedFieldPosition);
          END;
          (* ------------------------------ deallocate memory buffer *)
          Deallocate(modPtr, SIZE(ModInfo));
          (* ----------------------------------------- normal end *)
          ok := (linesPerPage >= Profiles[m].linesPerScreen);
        ELSE
          (* ---------------------- VESA mode isn't supported *)
          ok := FALSE;
        END;
        (* -------------------- deallocate memory buffer *)
        Deallocate(vgaPtr, SIZE(VgaInfo));
        (* ---------------- fail due to invalid mode *)
        testPassed := ok;
      ELSE
        (* -------------- VESA isn't supported *)
        testPassed := FALSE;
      END;
    END PrepareSuper;

  (* ================================ Initialization *)

  BEGIN
  (*%T EXTNO*)
    videoSel := ScreenSeg;
  (*%E*)
  (*%T EXTTS*)
    TSXLIB.InitInt10();
    videoSel := TSXLIB.SEL_A000H;
  (*%E*)
  END ScrExt.
