IMPLEMENTATION MODULE ScrNc8;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    *************************************************************************
    Hardware:   Video Graphics Array
    Compiler:   TopSpeed / Stony Brook Modula-2
    -------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Derived from modules RDEVGA4/RDCGA2    May 91
    A.Iakovlev  Ver  1.1  Corrections in RectCopy                June 91
    A.Iakovlev  Ver  1.2  Use combined low-level module RM       January 92
    A.Iakovlev  Ver  2.0  Module Res + procedures New/Dispose    August 92
    A.Iakovlev  Ver  3.0  InitPage/InitPageAddrs/TermPageAddrs   August 92
    A.Iakovlev  Ver  4.0  Renaming                               June 93
    A.Iakovlev  Ver  5.0  Import from SYSTEM                     January 94
    A.Iakovlev  Ver  5.1  Use SrvSys.Add instead of Lib.IncAddr  January 94
    A.Iakovlev  Ver  5.2  No use SHORTCARD type                  February 94
    A.Iakovlev  Ver  5.3  No shift operators                     February 94
    A.Iakovlev  Ver  5.4  Proc InitPage: width in bytes          February 94
    A.Iakovlev  Ver  6.0  Color type = Scr.Pixel                 September 94
    A.Iakovlev  Ver  6.1  Use module Mem for heap control        November 94
    *************************************************************************
    VGA hardware dependencies:
    ------ one byte per pixel, four planes for four sequential pixels
    ------ pixels positions on the pixmap
    ------ control via ports
    TopSpeed compiler dependencies:
    ------ options (*$...*)
    ------ structured constants
    ------ one-byte sets
    TopSpeed library dependencies:
    ------ import from SYSTEM: BYTE
    Stony Brook compiler dependencies:
    ------ structured constants
    ------ one-byte sets
    Stony Brook library dependencies:
    ------ import from SYSTEM: BYTE
  *)

  FROM Cfg IMPORT CMPTS, CMPSB;
  IMPORT
    SYSTEM,
    Geo, Res, DrvMem, SrvSys, SrvMat, SrvObj, SrvLow,
    ScrTyp, ScrVar, ScrLow, Scr;

  TYPE
    BS = ScrTyp.BS;
    BP = ScrTyp.BP;

  (* ========================= Basic constants *)

  CONST
    BitsPerPixel = 2;
    PlanesPerPixel = 4;
    PixelsPerByte = 8 DIV BitsPerPixel;

  (*
    NOTE: PixelsPerByte = PlanesPerPixel !!!!!!
  *)

  (* ========================================== Screen masks *)

  TYPE
    BM = ARRAY [0 .. PixelsPerByte - 1] OF BS;

  (*%T CMPTS*)

  CONST
    LftMask = BM(BS{0 .. 3}, BS{1 .. 3}, BS{2 .. 3}, BS{3     });
    RhtMask = BM(BS{0     }, BS{0 .. 1}, BS{0 .. 2}, BS{0 .. 3});
    (* ----------------------------------------------------------- *)
    ShlMask = BM(BS{0}, BS{1}, BS{2}, BS{3});

  (*%E*)

  (*%T CMPSB*)

  CONST
    LftMask:BM = [BS{0 .. 3}, BS{1 .. 3}, BS{2 .. 3}, BS{3     }];
    RhtMask:BM = [BS{0     }, BS{0 .. 1}, BS{0 .. 2}, BS{0 .. 3}];
    (* ----------------------------------------------------------- *)
    ShlMask:BM = [BS{0}, BS{1}, BS{2}, BS{3}];

  (*%E*)

  (* =========================================== Nibble invert masks *)

  CONST
    BitsPerNibble = 4;
    NibblesPerByte = 8 DIV BitsPerNibble;
    CombinationsPerNibble = 16; (* 1 << BitsPerNibble *)

  TYPE
    PCS = ARRAY [0 .. CombinationsPerNibble - 1] OF BS;

  (*%T CMPTS*)

  CONST
    InvMask = PCS
      (
        BS{    }, BS{      3}, BS{      2}, BS{      2, 3},
        BS{   1}, BS{   1, 3}, BS{   1, 2}, BS{   1, 2, 3},
        BS{0   }, BS{0,    3}, BS{0,    2}, BS{0,    2, 3},
        BS{0, 1}, BS{0, 1, 3}, BS{0, 1, 2}, BS{0, 1, 2, 3}
      );

  (*%E*)

  (*%T CMPSB*)

  CONST
    InvMask:PCS =
      [
        BS{    }, BS{      3}, BS{      2}, BS{      2, 3},
        BS{   1}, BS{   1, 3}, BS{   1, 2}, BS{   1, 2, 3},
        BS{0   }, BS{0,    3}, BS{0,    2}, BS{0,    2, 3},
        BS{0, 1}, BS{0, 1, 3}, BS{0, 1, 2}, BS{0, 1, 2, 3}
      ];

  (*%E*)

  (* ======================================= Line parameters *)

  TYPE
    LineParams =
      RECORD
        byteCount: CARDINAL;
        LM, RM: BS;
      END;

  (*$N,I-,O-*)

  PROCEDURE SetLineParams(rxMin, rxMax: CARDINAL; VAR lp: LineParams);
    BEGIN
      WITH lp DO
        byteCount := rxMax DIV PixelsPerByte - rxMin DIV PixelsPerByte;
        IF byteCount > 0 THEN
          LM := LftMask[rxMin MOD PixelsPerByte];
          RM := RhtMask[rxMax MOD PixelsPerByte];
        ELSE
          LM := LftMask[rxMin MOD PixelsPerByte] * RhtMask[rxMax MOD PixelsPerByte];
          RM := LM;
        END;
      END;
    END SetLineParams;

  (*$F,I=,O=*)

  (* ===================================== Full-pixmap functions *)

  PROCEDURE InitPage(w, h: CARDINAL);
    VAR
      bytesPerPage: CARDINAL;
      r: Geo.Rectangle;
    BEGIN
      bytesPerPage := w * h;
      Geo.SetRect(r, 0, (w * PixelsPerByte) - 1, 0, h - 1);
      WITH ScrVar.SPM DO
        number := MAX(CARDINAL) DIV bytesPerPage;
        length := bytesPerPage;
        width := w;
        surf := r;
        view.port := r;
      END;
    END InitPage;

  PROCEDURE InitPageAddrs(sel: CARDINAL);
    VAR
      pl: CARDINAL;
    BEGIN
      FOR pl := 0 TO PlanesPerPixel - 1 DO
        ScrVar.SPM.base[pl].addr := SrvSys.Addr(sel, 0);
      END;
    END InitPageAddrs;

  PROCEDURE TermPageAddrs();
    BEGIN
      ScrVar.SPM.base := ScrTyp.NilBanks;
    END TermPageAddrs;

  PROCEDURE New(r: Geo.Rectangle; VAR pm: PIXMAP): Res.Status;
    VAR
      newPm: PIXMAP;
      longLen: LONGCARD;
      lftByte: INTEGER;
      pl: CARDINAL;
    BEGIN
      IF DrvMem.Alloc(newPm, SIZE(ScrTyp.PixmapDescriptor)) THEN
        WITH newPm^ DO
          (* ------------------- compute width and length of plane *)
          lftByte := SrvMat.Div(r.ixMin, PixelsPerByte);
          width := SrvMat.Div(r.ixMax, PixelsPerByte) - lftByte + 1;
          longLen := VAL(LONGCARD, width) * VAL(LONGCARD, r.iyMax - r.iyMin + 1);
          (* ----------------------------------------- is the plane allocatable? *)
          IF longLen <= VAL(LONGCARD, MAX(CARDINAL)) THEN
            (* --------------------------------- set length of plane *)
            length := VAL(CARDINAL, longLen);
            (* ----------------------- attempt to allocate pixmap planes *)
            base := ScrTyp.NilBanks;
            pl := 0;
            LOOP
              IF pl < PlanesPerPixel THEN
                IF DrvMem.Alloc(base[pl].addr, length) THEN
                  (* ------------------ the plane is allocated *)
                  INC(pl);
                ELSE
                  (* ----------------- fail during plane allocation *)
                  EXIT;
                END;
              ELSE
                (* ----- all planes ok: attempt to allocate buffer *)
                IF DrvMem.Alloc(buffer.addr, width + 1) THEN
                  (* ------------------------------ set other parameters *)
                  number := PlanesPerPixel;
                  WHILE (length > MAX(CARDINAL) DIV number) DO
                    number := number DIV 2;
                  END;
                  surf := r;
                  ix0 := lftByte * PixelsPerByte;
                  iy0 := r.iyMin;
                  view.port := surf;
                  view.clip := TRUE;
                  screen := FALSE;
                  (* ---------- include into the pixmap stack *)
                  next := ScrVar.topPm;
                  ScrVar.topPm := newPm;
                  (* --------------- normal exit *)
                  pm := newPm;
                  RETURN Res.done;
                ELSE
                  (* ---------- fail during buffer allocation *)
                  EXIT;
                END;
              END;
            END;
            (* ---------------------------- fail due to heap limitations *)
            WHILE pl > 0 DO
              DEC(pl);
              DrvMem.Free(base[pl].addr, length);
            END;
            DrvMem.Free(newPm, SIZE(ScrTyp.PixmapDescriptor));
            RETURN Res.noMemory;
          ELSE
            (* ------------------ fail: too large pixmap requested *)
            DrvMem.Free(newPm, SIZE(ScrTyp.PixmapDescriptor));
            RETURN Res.tooLargeRectangle;
          END;
        END;
      ELSE
        RETURN Res.noMemory;
      END;
    END New;

  PROCEDURE Dispose(VAR pm: PIXMAP);
    VAR
      prevPm: PIXMAP;
      pl: CARDINAL;
    BEGIN
      prevPm := PIXMAP(SrvObj.Prev(SrvObj.OBJECT(ScrVar.topPm), SrvObj.OBJECT(pm)));
      IF prevPm # NIL THEN
        WITH pm^ DO
          IF NOT screen THEN
            (* ----------- deallocate the pixmap content and buffer *)
            DrvMem.Free(buffer.addr, width + 1);
            pl := PlanesPerPixel;
            WHILE pl > 0 DO
              DEC(pl);
              DrvMem.Free(base[pl].addr, length);
            END;
            (* ---------------------------- exclude from the pixmap stack *)
            IF pm # ScrVar.topPm THEN
              prevPm^.next := next;
            ELSE
              ScrVar.topPm := next;
            END;
            (* ---------------------------- deallocate the descriptor *)
            DrvMem.Free(pm, SIZE(ScrTyp.PixmapDescriptor));
          END;
        END;
      END;
    END Dispose;

  PROCEDURE Copy(dst, src: PIXMAP);
    VAR
      pl: CARDINAL;
    BEGIN
      WITH dst^ DO
        IF screen THEN
          IF src^.screen THEN
            (* ---------- copy screen to screen: set write mode *)
            ScrLow.VGASetWMode1();
            (* --------------------------------------- use byte moves! *)
            SrvLow.MoveInc(src^.base[0].addr, base[0].addr, length);
            (* ----------------------------------- reset write mode *)
            ScrLow.VGAResetMode();
          ELSE
            (* --------------------- copy memory to screen *)
            FOR pl := 0 TO PlanesPerPixel - 1 DO
              (* -------------------------------- set map mask *)
              ScrLow.SetMapMask(ShlMask[pl]);
              (* -------------------------------------- use byte moves! *)
              SrvLow.MoveInc(src^.base[pl].addr, base[0].addr, length);
            END;
            (* --------------------------- reset map mask *)
            ScrLow.ResetMapMask();
          END;
        ELSE
          IF src^.screen THEN
            (* ------------------------ copy screen to memory *)
            FOR pl := 0 TO PlanesPerPixel - 1 DO
              (* ---------------------------------- set read map *)
              ScrLow.SetReadMap(SYSTEM.BYTE(pl));
              (* -------------------------------------- use byte moves! *)
              SrvLow.MoveInc(src^.base[0].addr, base[pl].addr, length);
            END;
          ELSE
            (* --------------- copy memory to memory using fast (word) moves *)
            FOR pl := 0 TO PlanesPerPixel - 1 DO
              SrvLow.FastMoveInc(src^.base[pl].addr, base[pl].addr, length);
            END;
          END;
        END;
      END;
    END Copy;

  (* ============================================= Graphics primitives *)

  PROCEDURE Clear(pm: PIXMAP; c: Scr.Pixel);
    VAR
      pl: CARDINAL;
    BEGIN
      WITH pm^ DO
        IF screen THEN
          (* ----------------------- clear on the screen *)
          SrvLow.Fill(base[0].addr, length, BS(c));
        ELSE
          (* --------------------------- clear in the memory *)
          FOR pl := 0 TO PlanesPerPixel - 1 DO
            SrvLow.FastFill(base[pl].addr, length, BS(c));
          END;
        END;
      END;
    END Clear;

  PROCEDURE Pt(pm: PIXMAP; p: Geo.Point; c: Scr.Pixel; func: Scr.Function);
    VAR
      ptr: ScrTyp.BP;
  (*$W+*)
      TM: BS;
  (*$W=*)
    BEGIN
      WITH pm^ DO
        (* ------------------------------------------- clipping *)
        IF view.clip AND NOT Geo.PtInRect(p, view.port) THEN
          RETURN;
        END;
        (* ------------------------------------- drawing ... *)
        IF screen THEN
          (* -------- ... on the screen: set the destination pointer *)
          ptr := base[0];
          INC(ptr.ofs,
            VAL(CARDINAL, p.iy) * width +
            VAL(CARDINAL, p.ix) DIV PixelsPerByte);
          (* --------------------------------------- set registers *)
          ScrLow.SetFuncMapMask
            (
              SYSTEM.BYTE(func),
              ShlMask[VAL(CARDINAL, p.ix) MOD PixelsPerByte]
            );
          (* ----------------------------------- set the byte *)
          TM := ptr.addr^;
          ptr.addr^ := BS(c);
          (* ------------------------- reset registers *)
          ScrLow.ResetFuncMapMask();
        ELSE
          (* ------------- ... in the memory: set the destination pointer *)
          ptr := base[VAL(CARDINAL, p.ix - ix0) MOD PixelsPerByte];
          INC(ptr.ofs,
            VAL(CARDINAL, p.iy - iy0) * width +
            VAL(CARDINAL, p.ix - ix0) DIV PixelsPerByte);
          (* -------------------------------------- set the byte *)
          IF func = Scr.copy THEN
            ptr.addr^ := BS(c);
          ELSE
            ptr.addr^ := ptr.addr^ / BS(c);
          END;
        END;
      END;
    END Pt;

  PROCEDURE PtColor(pm: PIXMAP; p: Geo.Point): Scr.Pixel;
    VAR
      ptr: ScrTyp.BP;
    BEGIN
      (* --------------- set the parameters *)
      WITH pm^ DO
        IF screen THEN
          (* ------------------ the screen source pointer *)
          ptr := base[0];
          INC(ptr.ofs,
            VAL(CARDINAL, p.iy) * width +
            VAL(CARDINAL, p.ix) DIV PixelsPerByte);
          (* ------------------------------------------------ the read map *)
          ScrLow.SetReadMap(SYSTEM.BYTE(VAL(CARDINAL, p.ix) MOD PixelsPerByte));
        ELSE
          (* ---------------------------- the memory source pointer *)
          ptr := base[VAL(CARDINAL, p.ix - ix0) MOD PixelsPerByte];
          INC(ptr.ofs,
            VAL(CARDINAL, p.iy - iy0) * width +
            VAL(CARDINAL, p.ix - ix0) DIV PixelsPerByte);
        END;
      END;
      (* -------------------------- return the result *)
      RETURN VAL(Scr.Pixel, ORD(CHAR(ptr.addr^)));
    END PtColor;

  PROCEDURE SolidRect(pm: PIXMAP; r: Geo.Rectangle; c: Scr.Pixel; func: Scr.Function);
    VAR
      lp: LineParams;
      pl, byteCount, lineCount, lineOfs, line, ofs1: CARDINAL;
      ptr: ScrTyp.BP;
  (*$W+*)
      TM: BS;
  (*$W=*)
    BEGIN
      WITH pm^ DO
        (* ---------------------------------------------------- clipping *)
        IF view.clip AND NOT Geo.RectXRectRes(r, view.port, r) THEN
          RETURN;
        END;
        (* ---------------------------------------- set line parameters *)
        SetLineParams(r.ixMin - ix0, r.ixMax - ix0, lp);
        (* --------------------------------------- set the line count *)
        lineCount := r.iyMax - r.iyMin;
        (* -------------------------------------------- drawing ... *)
        IF screen THEN
          (* ------ ... on the screen: set the destination pointer *)
          ptr := base[0];
          INC(ptr.ofs,
            VAL(CARDINAL, r.iyMin) * width +
            VAL(CARDINAL, r.ixMin) DIV PixelsPerByte);
          (* ------------------------------- set the line offset *)
          lineOfs := width - lp.byteCount;
          (* ---------------------------------------- set registers *)
          ScrLow.SetFuncMapMask(SYSTEM.BYTE(func), SYSTEM.BYTE(lp.LM));
          (* ----------------------------------------------- main loop *)
          line := 0;
          LOOP
            (* --------------------------------- set the left byte *)
            TM := ptr.addr^;
            ptr.addr^ := BS(c);
            (* ---------------------------------------- set other bytes *)
            IF lp.byteCount > 0 THEN
              (* ----------------------------------- increment the offset *)
              INC(ptr.ofs);
              (* --------------------------------------- set inner bytes *)
              IF lp.byteCount > 1 THEN
                ScrLow.ResetMapMask();
                IF func = Scr.copy THEN
                  SrvLow.Fill(ptr.addr, lp.byteCount - 1, BS(c));
                ELSE
                  SrvLow.Refill(ptr.addr, lp.byteCount - 1, BS(c));
                END;
                INC(ptr.ofs, lp.byteCount - 1);
              END;
              (* ----------------------------- set the right byte *)
              ScrLow.SetMapMask(SYSTEM.BYTE(lp.RM));
              TM := ptr.addr^;
              ptr.addr^ := BS(c);
              (* ------------ reset the mask for the left byte *)
              ScrLow.SetMapMask(SYSTEM.BYTE(lp.LM));
            END;
            (* ------------------- check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ------------------- increment the offset *)
            INC(ptr.ofs, lineOfs);
            (* --------- increment the line counter *)
            INC(line);
          END;
          (* ---------------------------- reset registers *)
          ScrLow.ResetFuncMapMask();
        ELSE
          (* ------ ... in the memory: set the constant part of the offset *)
          ofs1 :=
            VAL(CARDINAL, r.iyMin - iy0) * width +
            VAL(CARDINAL, r.ixMin - ix0) DIV PixelsPerByte;
          (* --------------------------------------------- main loop *)
          line := 0;
          LOOP
            (* ---------------------- loop through all the planes *)
            FOR pl := 0 TO PlanesPerPixel - 1 DO
              (* ---------------------------------- check bytes count *)
              IF lp.byteCount > 0 THEN
                (* ---- wide rectangle: set the pointer and bytes count *)
                ptr := base[pl];
                INC(ptr.ofs, ofs1);
                byteCount := lp.byteCount;
                (* -------------------- correct the pointer and bytes count *)
                IF NOT (pl IN lp.LM) THEN INC(ptr.ofs); DEC(byteCount) END;
                IF pl IN lp.RM THEN INC(byteCount) END;
                (* ---------------------------------- check bytes count again *)
                IF byteCount > 0 THEN
                  (* ----------------- check the function and output *)
                  IF func = Scr.copy THEN
                    SrvLow.FastFill(ptr.addr, byteCount, BS(c));
                  ELSE
                    SrvLow.FastFillXor(ptr.addr, byteCount, BS(c));
                  END;
                END;
              ELSIF pl IN lp.LM THEN
                (* ------- narrow rectangle: set the destination pointer *)
                ptr := base[pl];
                INC(ptr.ofs, ofs1);
                (* ---------- check the function and output *)
                IF func = Scr.copy THEN
                  ptr.addr^ := BS(c);
                ELSE
                  ptr.addr^ := ptr.addr^ / BS(c);
                END;
              END;
            END;
            (* ------------------ check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ----------------- increment the offset *)
            INC(ofs1, width);
            (* --------- increment the line counter *)
            INC(line);
          END;
        END;
      END;
    END SolidRect;

  PROCEDURE StippledRect(pm: PIXMAP; r: Geo.Rectangle;
            stipple: SYSTEM.ADDRESS; w, h: CARDINAL; p: Geo.Point;
            c: Scr.Pixel; func: Scr.Function);

    VAR
      lp: LineParams;
      pl, byte, line, lineCount, lineOfs, ofs1: CARDINAL;
      srcOfs0, srcOfs1, srcByteCount: CARDINAL;
      stippleOfs0, stippleLineOfs, deltaLow, deltaHgh: CARDINAL;
      ptr, srcPtr, stipplePtr: ScrTyp.BP;
      stippleLarge, shiftNeed: BOOLEAN;
  (*$W+*)
      TM: BS;
  (*$W=*)

    PROCEDURE ComputeSrc();
      TYPE
        DoubleStipple =
          RECORD
            CASE :BOOLEAN OF
            | TRUE:
              card: CARDINAL;
            | FALSE:
              low, hgh: BS;
            END;
          END;
      VAR
        stippleBuffer: DoubleStipple;
        stippleByte: CARDINAL;
      BEGIN
        IF shiftNeed THEN
          (* ------- preread last stipple byte for first source stipple *)
          IF stippleLarge THEN
            INC(stipplePtr.ofs, w - 1);
            stippleBuffer.low := stipplePtr.addr^;
            DEC(stipplePtr.ofs, w - 1);
          ELSE
            stippleBuffer.low := stipplePtr.addr^;
          END;
          (* -------------------------------- compute stipples with shifting *)
          FOR stippleByte := 0 TO w - 1 DO
            stippleBuffer.hgh := stippleBuffer.low;
            stippleBuffer.low := stipplePtr.addr^;
            INC(stipplePtr.ofs);
            srcPtr.addr^ := InvMask[SrvLow.Shr(stippleBuffer.card, deltaHgh) MOD CombinationsPerNibble];
            INC(srcPtr.ofs);
            srcPtr.addr^ := InvMask[SrvLow.Shr(stippleBuffer.card, deltaLow) MOD CombinationsPerNibble];
            INC(srcPtr.ofs);
          END;
        ELSE
          (* ------------------ compute stipples without shifting *)
          FOR stippleByte := 0 TO w - 1 DO
            stippleBuffer.low := stipplePtr.addr^;
            INC(stipplePtr.ofs);
            srcPtr.addr^ := InvMask[SrvLow.Shr(stippleBuffer.card, BitsPerNibble) MOD CombinationsPerNibble];
            INC(srcPtr.ofs);
            srcPtr.addr^ := InvMask[(stippleBuffer.card) MOD CombinationsPerNibble];
            INC(srcPtr.ofs);
          END;
        END;
      END ComputeSrc;

    PROCEDURE IncSrcOfs();
      BEGIN
        IF stippleLarge THEN
          INC(srcPtr.ofs);
          IF (srcPtr.ofs - srcOfs0) MOD srcByteCount = 0 THEN
            DEC(srcPtr.ofs, srcByteCount);
          END;
        ELSIF srcPtr.ofs = srcOfs0 THEN
          INC(srcPtr.ofs);
        ELSE
          DEC(srcPtr.ofs);
        END;
      END IncSrcOfs;

    PROCEDURE SetPt();
      BEGIN
        IF func = Scr.copy THEN
          ptr.addr^ := BS(c);
        ELSE
          ptr.addr^ := ptr.addr^ / BS(c);
        END;
      END SetPt;

    BEGIN
      WITH pm^ DO
        (* ------------------------------------------------------ clipping *)
        IF view.clip AND NOT Geo.RectXRectRes(r, view.port, r) THEN
          RETURN;
        END;
        (* -------------------------------------- set line parameters *)
        SetLineParams(r.ixMin - ix0, r.ixMax - ix0, lp);
        (* ----------------------------------- set the line count *)
        lineCount := r.iyMax - r.iyMin;
        (* -------------------------------- set stipple parameters *)
        stipplePtr.addr := stipple;
        stippleOfs0 := stipplePtr.ofs;
        stippleLineOfs := SrvMat.Mod(r.iyMin - p.iy, h);
        stippleLarge := w > 1;
        (* ------------------------------- set shift parameters *)
        deltaLow := SrvMat.Mod(p.ix, PixelsPerByte);
        shiftNeed := deltaLow > 0;
        IF shiftNeed THEN
          deltaHgh := deltaLow + BitsPerNibble;
        END;
        (* ----------------------------------- set source parameters *)
        srcByteCount := w * NibblesPerByte;
        srcOfs0 := buffer.ofs;
        srcOfs1 :=
          SrvMat.Mod(SrvMat.Div(r.ixMin - p.ix, PixelsPerByte), srcByteCount) +
          srcOfs0;
        (* -------------------------------------------------- drawing ... *)
        IF screen THEN
          (* --------- ... on the screen: set the destination pointer *)
          ptr := base[0];
          INC(ptr.ofs,
            VAL(CARDINAL, r.iyMin) * width +
            VAL(CARDINAL, r.ixMin) DIV PixelsPerByte);
          (* -------------------------- set the line offset *)
          lineOfs := width - lp.byteCount;
          (* ----------------------------- set registers *)
          ScrLow.SetFunc(SYSTEM.BYTE(func));
          (* -------------------------------- main loop *)
          line := 0;
          LOOP
            (* ------------------------ set the stipple offset *)
            stipplePtr.ofs := stippleLineOfs * w + stippleOfs0;
            (* ----------------------------------- check the stipple *)
            IF (BS(stipplePtr.addr^) # BS{}) OR stippleLarge THEN
              (* --- stipple is nonempty or large: set the source pointer *)
              srcPtr := buffer;
              (* ------------------------------------- compute the source *)
              ComputeSrc();
              (* ---------------------------------- set the source offset *)
              srcPtr.ofs := srcOfs1;
              (* ----------------------------------- set the left byte *)
              ScrLow.SetMapMask(SYSTEM.BYTE(srcPtr.addr^ * lp.LM));
              TM := ptr.addr^;
              ptr.addr^ := BS(c);
              (* --------------------------------- set other bytes *)
              IF lp.byteCount > 0 THEN
                (* ----------------- increment pointer offsets *)
                INC(ptr.ofs);
                IncSrcOfs();
                (* ------------------------ set inner bytes *)
                IF lp.byteCount > 1 THEN
                  byte := 2;
                  LOOP
                    ScrLow.SetMapMask(SYSTEM.BYTE(srcPtr.addr^));
                    TM := ptr.addr^;
                    ptr.addr^ := BS(c);
                    INC(ptr.ofs);
                    IncSrcOfs();
                    IF byte = lp.byteCount THEN EXIT END;
                    INC(byte);
                  END;
                END;
                (* --------------------------------- set the right byte *)
                ScrLow.SetMapMask(SYSTEM.BYTE(srcPtr.addr^ * lp.RM));
                TM := ptr.addr^;
                ptr.addr^ := BS(c);
              END;
            ELSE
              (* --- stipple is empty and not-large: increment the offset only *)
              INC(ptr.ofs, lp.byteCount);
            END;
            (* ----------------------------- check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ----------- increment the destination offset *)
            INC(ptr.ofs, lineOfs);
            (* --------------- increment the stipple line *)
            stippleLineOfs := (stippleLineOfs + 1) MOD h;
            (* --------- increment the line counter *)
            INC(line);
          END;
          (* ------------------------------ reset registers *)
          ScrLow.ResetFuncMapMask();
        ELSE
          (* ---- ... in the memory: set the constant part of the offset *)
          ofs1 :=
            VAL(CARDINAL, r.iyMin - iy0) * width +
            VAL(CARDINAL, r.ixMin - ix0) DIV PixelsPerByte;
          (* ----------------------------------------- main loop *)
          line := 0;
          LOOP
            (* ------------------------- set the stipple offset *)
            stipplePtr.ofs := stippleLineOfs * w + stippleOfs0;
            (* ---------------------------------- check the stipple *)
            IF (BS(stipplePtr.addr^) # BS{}) OR stippleLarge THEN
              (* --- stipple is nonempty or large: set the source pointer *)
              srcPtr := buffer;
              (* ------------------------------------ compute the source *)
              ComputeSrc();
              (* ----------------------------- loop through all the planes *)
              FOR pl := 0 TO PlanesPerPixel - 1 DO
                (* ------------------------------ set the source offset *)
                srcPtr.ofs := srcOfs1;
                (* ------------- set the destination pointer *)
                ptr := base[pl];
                INC(ptr.ofs, ofs1);
                (* -------------------------------- set the left byte *)
                IF pl IN (srcPtr.addr^ * lp.LM) THEN SetPt() END;
                (* -------------------------------------- set other bytes *)
                IF lp.byteCount > 0 THEN
                  (* ----------------------------- increment pointer offsets *)
                  INC(ptr.ofs);
                  IncSrcOfs();
                  (* -------------------------------------- set inner bytes *)
                  IF lp.byteCount > 1 THEN
                    byte := 2;
                    LOOP
                      IF pl IN srcPtr.addr^ THEN SetPt() END;
                      INC(ptr.ofs);
                      IncSrcOfs();
                      IF byte = lp.byteCount THEN EXIT END;
                      INC(byte);
                    END;
                  END;
                  (* -------------------------------- set the right byte *)
                  IF pl IN (srcPtr.addr^ * lp.RM) THEN SetPt() END;
                END;
              END;
            END;
            (* ---------------------------- check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ------------------ increment the destination offset *)
            INC(ofs1, width);
            (* --------------------- increment the stipple line *)
            stippleLineOfs := (stippleLineOfs + 1) MOD h;
            (* --------------- increment the line counter *)
            INC(line);
          END;
        END;
      END;
    END StippledRect;

  TYPE CS = ARRAY [0 .. PlanesPerPixel] OF CARDINAL;

  (*%T CMPTS*)
  CONST NmbOfContours = CS(4, 1, 2, 1, 4);
  (*%E*)

  (*%T CMPSB*)
  CONST NmbOfContours:CS = [4, 1, 2, 1, 4];
  (*%E*)

  PROCEDURE RectCopy(dst: PIXMAP; r: Geo.Rectangle;
            src: PIXMAP; p: Geo.Point; func: Scr.Function);
    VAR
      lp: LineParams;
      ox, oy, rxBegin, ryBegin, pxBegin, pyBegin: INTEGER;
      contNo, pl, srcPl, byteCount, line, lineCount, ofs1, srcOfs1, delta: CARDINAL;
      forwardX, forwardY, buffNeed: BOOLEAN;
      ptr, srcPtr, prePtr, buffPtr: ScrTyp.BP;
      SM: BS;
  (*$W+*)
      TM: BS;
  (*$W=*)
    BEGIN
      WITH dst^ DO
        (* ----------------------------- clipping *)
        IF view.clip THEN
          ox := r.ixMin;
          oy := r.iyMin;
          IF Geo.RectXRectRes(r, view.port, r) THEN
            (* ------------------- shift of the source base point *)
            INC(p.ix, r.ixMin - ox);
            INC(p.iy, r.iyMin - oy);
          ELSE
            (* ------------------------- empty intersection *)
            RETURN;
          END;
        END;
        (* ------------------------------- set line parameters *)
        SetLineParams(r.ixMin - ix0, r.ixMax - ix0, lp);
        (* ---------------------------------- set the line count *)
        lineCount := r.iyMax - r.iyMin;
        (* ------------------------------------ set direction flags *)
        forwardX := p.ix >= r.ixMin;
        forwardY := p.iy >= r.iyMin;
        (* ------------ set parameters according with X-direction flag *)
        IF forwardX THEN
          (* ----------------------------------- ... beginning corners *)
          rxBegin := r.ixMin;
          pxBegin := p.ix;
          (* ---------------------------------------- ... shift parameters *)
          delta := VAL(CARDINAL, pxBegin - rxBegin) MOD PixelsPerByte;
          IF delta # 0 THEN
            SM := LftMask[SrvMat.Mod(pxBegin, PixelsPerByte)];
          ELSE
            SM := lp.LM;
          END;
        ELSE
          (* ----------------------------------- ... beginning corners *)
          rxBegin := r.ixMax;
          pxBegin := p.ix + r.ixMax - r.ixMin;
          (* ---------------------------------------- ... shift parameters *)
          delta := PixelsPerByte - VAL(CARDINAL, rxBegin - pxBegin) MOD PixelsPerByte;
          IF delta # PixelsPerByte THEN
            SM := RhtMask[SrvMat.Mod(pxBegin, PixelsPerByte)];
          ELSE
            SM := lp.RM;
          END;
        END;
        (* ----------- set parameters according with Y-direction flag *)
        IF forwardY THEN
          ryBegin := r.iyMin;
          pyBegin := p.iy;
        ELSE
          ryBegin := r.iyMax;
          pyBegin := p.iy + INTEGER(lineCount);
        END;
        (* -------------------------------------------------- copying *)
        IF
          screen AND
          src^.screen AND
          (lp.LM = BS{0 .. PixelsPerByte - 1}) AND
          (NOT (lp.byteCount > 0) OR (lp.RM = BS{0 .. PixelsPerByte - 1})) AND
          (func = Scr.copy) AND
          (delta MOD PixelsPerByte = 0)
        THEN
          (* -------- copy in write mode 1: set the destination pointer *)
          ptr := base[0];
          INC(ptr.ofs,
            VAL(CARDINAL, ryBegin) * width +
            VAL(CARDINAL, rxBegin) DIV PixelsPerByte);
          (* ----------------- set the source pointer: just the same width! *)
          srcPtr := src^.base[0];
          INC(srcPtr.ofs,
            VAL(CARDINAL, pyBegin) * width +
            VAL(CARDINAL, pxBegin) DIV PixelsPerByte);
          (* ------------------------------------------------ set registers *)
          ScrLow.VGASetWMode1();
          (* --------------------------------------------------- main loop *)
          line := 0;
          LOOP
            (* -------------------------- copy the line using byte moves *)
            IF forwardX THEN
              SrvLow.MoveInc(srcPtr.addr, ptr.addr, lp.byteCount + 1);
            ELSE
              SrvLow.MoveDec(srcPtr.addr, ptr.addr, lp.byteCount + 1);
            END;
            (* --------------------------- check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* ---------------------------------- change offsets *)
            IF forwardY THEN
              INC(ptr.ofs, width);
              INC(srcPtr.ofs, width);
            ELSE
              DEC(ptr.ofs, width);
              DEC(srcPtr.ofs, width);
            END;
            (* ------------ increment the line counter *)
            INC(line);
          END;
          (* ------------------------- reset the write mode *)
          ScrLow.VGAResetMode();
        ELSE
          (* ------- copy in write mode 0: set the destination offset *)
          ofs1 :=
            VAL(CARDINAL, ryBegin - iy0) * width +
            VAL(CARDINAL, rxBegin - ix0) DIV PixelsPerByte;
          (* ----------------------------- set the source offset *)
          WITH src^ DO
            srcOfs1 :=
              VAL(CARDINAL, pyBegin - iy0) * width +
              VAL(CARDINAL, pxBegin - ix0) DIV PixelsPerByte;
          END;
          (* ------------------ set registers iff need *)
          IF screen THEN
            ScrLow.SetFunc(SYSTEM.BYTE(func));
          END;
          (* --------------------------------- set buffering parameters *)
          buffNeed := (dst = src) AND (delta MOD PixelsPerByte # 0);
          IF buffNeed THEN
            buffPtr := buffer;
            IF NOT forwardX THEN INC(buffPtr.ofs, lp.byteCount + 1) END;
          END;
          (* ------------------------------------------ main loop *)
          line := 0;
          LOOP
            (* ----------------------- loop through all the contours *)
            FOR contNo := 0 TO NmbOfContours[ORD(delta)] - 1 DO
              (* ---------------- set the initial plane of the contour *)
              pl := contNo;
              (* ------------------------ buffering of the initial plane *)
              IF buffNeed THEN
                (* --------------- select the screen source plane iff need *)
                IF screen THEN
                  ScrLow.SetReadMap(SYSTEM.BYTE(pl));
                END;
                (* ------------------------------ set the prereading pointer *)
                prePtr := base[pl];
                INC(prePtr.ofs, srcOfs1);
                (* ----------------- preread the initial plane of the contour *)
                IF forwardX THEN
                  SrvLow.MoveInc(prePtr.addr, buffPtr.addr, lp.byteCount + 1);
                ELSE
                  SrvLow.MoveDec(prePtr.addr, buffPtr.addr, lp.byteCount + 1);
                END;
              END;
              (* ------------ loop through all the planes of the contour *)
              REPEAT
                (* -------------------------- select the destination plane *)
                IF screen THEN
                  ScrLow.SetMapMask(ShlMask[pl]);
                END;
                (* ------------- set the destination pointer and bytes count *)
                ptr := base[pl];
                INC(ptr.ofs, ofs1);
                byteCount := lp.byteCount;
                (* ---------------------------------- set the source plane *)
                srcPl := (pl + delta) MOD PlanesPerPixel;
                (* ------------------------------- set the source pointer *)
                IF buffNeed AND (srcPl = contNo) THEN
                  (* ------------------------------- equal to the buffer *)
                  srcPtr := buffPtr;
                ELSE
                  (* ---------- select the screen source plane iff need *)
                  IF src^.screen THEN
                    ScrLow.SetReadMap(SYSTEM.BYTE(srcPl));
                  END;
                  (* ------------- equal to the normal source pointer *)
                  srcPtr := src^.base[srcPl];
                  INC(srcPtr.ofs, srcOfs1);
                END;
                (* ---------------------- check the X-direction flag *)
                IF forwardX THEN
                  (* ---------- in forward X-direction: check bytes count *)
                  IF byteCount > 0 THEN
                    (* ------- wide rectangle: correct destination parameters *)
                    IF NOT (pl IN lp.LM) THEN INC(ptr.ofs); DEC(byteCount) END;
                    IF pl IN lp.RM THEN INC(byteCount) END;
                    (* ----------------------------- check bytes count again *)
                    IF byteCount > 0 THEN
                      (* ------------------- correct source offset iff need *)
                      IF NOT (srcPl IN SM) THEN INC(srcPtr.ofs) END;
                      (* ---------- copy depending on the function *)
                      IF func = Scr.copy THEN
                        SrvLow.MoveInc(srcPtr.addr, ptr.addr, byteCount);
                      ELSIF screen THEN
                        SrvLow.RemoveInc(srcPtr.addr, ptr.addr, byteCount);
                      ELSE
                        SrvLow.MoveXorInc(srcPtr.addr, ptr.addr, byteCount);
                      END;
                    END;
                  ELSIF pl IN lp.LM THEN
                    (* --- narrow rectangle: correct source offset iff need *)
                    IF NOT (srcPl IN SM) THEN INC(srcPtr.ofs) END;
                    (* ---------- copy depending on the function *)
                    IF func = Scr.copy THEN
                      ptr.addr^ := srcPtr.addr^;
                    ELSIF screen THEN
                      TM := ptr.addr^;
                      ptr.addr^ := srcPtr.addr^;
                    ELSE
                      ptr.addr^ := ptr.addr^ / srcPtr.addr^;
                    END;
                  END;
                ELSE
                  (* ---------- in backward X-direction: check bytes count *)
                  IF byteCount > 0 THEN
                    (* -------------- wide rectangle: correct copy parameters *)
                    IF NOT (pl IN lp.RM) THEN DEC(ptr.ofs); DEC(byteCount) END;
                    IF pl IN lp.LM THEN INC(byteCount) END;
                    (* ------------------------------ check byte count again *)
                    IF byteCount > 0 THEN
                      (* ------------------ correct source offset iff need *)
                      IF NOT (srcPl IN SM) THEN DEC(srcPtr.ofs) END;
                      (* -------- copy depending on the function *)
                      IF func = Scr.copy THEN
                        SrvLow.MoveDec(srcPtr.addr, ptr.addr, byteCount);
                      ELSIF screen THEN
                        SrvLow.RemoveDec(srcPtr.addr, ptr.addr, byteCount);
                      ELSE
                        SrvLow.MoveXorDec(srcPtr.addr, ptr.addr, byteCount);
                      END;
                    END;
                  ELSIF pl IN lp.RM THEN
                    (* --- narrow rectangle: correct source offset iff need *)
                    IF NOT (srcPl IN SM) THEN DEC(srcPtr.ofs) END;
                    (* ------------- copy depending on the function *)
                    IF func = Scr.copy THEN
                      ptr.addr^ := srcPtr.addr^;
                    ELSIF screen THEN
                      TM := ptr.addr^;
                      ptr.addr^ := srcPtr.addr^;
                    ELSE
                      ptr.addr^ := ptr.addr^ / srcPtr.addr^;
                    END;
                  END;
                END;
                (* ------------- set the next destination plane *)
                pl := srcPl;
                (* --------- check the stop condition *)
              UNTIL pl = contNo;
            END;
            (* --------------------------- check the stop condition *)
            IF line = lineCount THEN EXIT END;
            (* --------------------------------- change offsets *)
            IF forwardY THEN
              INC(ofs1, width);
              INC(srcOfs1, src^.width);
            ELSE
              DEC(ofs1, width);
              DEC(srcOfs1, src^.width);
            END;
            (* ---------- increment the line counter *)
            INC(line);
          END;
          (* ----------------------------- reset registers *)
          IF screen THEN
            ScrLow.ResetFuncMapMask();
          END;
        END;
      END;
    END RectCopy;

  (* ================================= Partitions store/restore functions *)

  PROCEDURE PartsNmb(pm: PIXMAP): CARDINAL;
    BEGIN
      WITH pm^ DO
        RETURN PlanesPerPixel DIV number;
      END;
    END PartsNmb;

  PROCEDURE PartSize(pm: PIXMAP; n: CARDINAL): CARDINAL;
    BEGIN
      WITH pm^ DO
        RETURN length * number;
      END;
    END PartSize;

  PROCEDURE Store(pm: PIXMAP; n: CARDINAL; s: SYSTEM.ADDRESS);
    VAR
      pl: CARDINAL;
    BEGIN
      WITH pm^ DO
        (* ------------------------- check partition number *)
        n := n MOD (PlanesPerPixel DIV number);
        (* ----------------------- store partition planes using byte moves *)
        FOR pl := n * number TO (n + 1) * number - 1 DO
          IF screen THEN ScrLow.SetReadMap(SYSTEM.BYTE(pl)) END;
          SrvLow.MoveInc(base[pl].addr, s, length);
          s := SrvSys.Add(s, length);
        END;
      END;
    END Store;

  PROCEDURE Restore(pm: PIXMAP; n: CARDINAL; s: SYSTEM.ADDRESS);
    VAR
      pl: CARDINAL;
    BEGIN
      WITH pm^ DO
        (* ----------------------------- check partition number *)
        n := n MOD (PlanesPerPixel DIV number);
        (* ----------------- restore partition planes using byte moves *)
        FOR pl := n * number TO (n + 1) * number - 1 DO
          IF screen THEN ScrLow.SetMapMask(ShlMask[pl]) END;
          SrvLow.MoveInc(s, base[pl].addr, length);
          s := SrvSys.Add(s, length);
        END;
        (* ----------------------------- restore registers *)
        IF screen THEN ScrLow.ResetMapMask() END;
      END;
    END Restore;

  END ScrNc8.
